@@ -53,6 +53,7 @@ function reboot() {
 	boot()
 	initFieldHolders()
 	initUserAgents()
+	initWarningsOnly()
 	installDynamicStyle()
 	updateFragmentIdHilite()
 	window.setInterval(emulateHashChanged, 50)
@@ -61,6 +62,7 @@ function reboot() {
 	injectHyperlinks()
 	moveLangAndDirWarningsAndAddLinks()
 	replaceSuccessFailure()
+	hideSourceIfNotRequested()
 }
 
 function installDynamicStyle() {
@@ -193,6 +195,25 @@ function toggleExtraOptions() {
 	}
 }
 
+function initWarningsOnly() {
+	var warningsCheckbox = document.getElementById("level")
+	if (!warningsCheckbox) {
+		return
+	}
+	if (supportsLocalStorage() && localStorage["warningsOnly"] == "yes") {
+		warningsCheckbox.checked = true
+	}
+	warningsCheckbox.addEventListener("change", function(e) {
+		if (supportsLocalStorage()) {
+			if (e.target.checked) {
+				localStorage["warningsOnly"] = "yes"
+			} else {
+				localStorage["warningsOnly"] = "no"
+			}
+		}
+	}, false)
+}
+
 function initUserAgents() {
 	var userAgents = document.querySelector("#useragents")
 	userAgents.appendChild(createLabeledOption(
@@ -321,6 +342,9 @@ function injectHyperlinks() {
 	linkify(errors, "all image candidate strings must specify a width",
 		"https://ericportis.com/posts/2014/srcset-sizes/",
 		"srcset and sizes overview")
+	linkify(info, "https://www.w3.org/International/articles/ruby/markup.en.html#visual",
+		"https://www.w3.org/International/articles/ruby/markup.en.html#visual",
+		"W3C guidance on ruby markup")
 	}
 
 function replaceSuccessFailure() {
@@ -440,7 +464,7 @@ function installTextarea() {
 			if (!document.getElementById("csslabel")) {
 				var cssLabel = document.createElement("label")
 				cssLabel.setAttribute("id", "csslabel")
-				cssLabel.setAttribute("title", "Treat the input as CSS.")
+				cssLabel.setAttribute("title", "Check the text input as CSS, not as HTML.")
 				cssLabel.setAttribute("for", "css")
 				var cssCheckbox = document.createElement("input")
 				cssCheckbox.setAttribute("type", "checkbox")
@@ -468,7 +492,7 @@ function installTextarea() {
 					}
 				}, false)
 				cssLabel.appendChild(cssCheckbox)
-				cssLabel.appendChild(document.createTextNode("CSS"))
+				cssLabel.appendChild(document.createTextNode("check as CSS"))
 				inputRegion.appendChild(cssLabel)
 			}
 			inputRegion.appendChild(textarea)
@@ -481,10 +505,6 @@ function installTextarea() {
 			submit.parentNode.appendChild(textareaHidden)
 		}
 	}
-	var showSource = document.getElementById("showsource")
-	if (showSource) {
-		showSource.checked = true
-	}
 }
 
 function installFileUpload() {
@@ -560,6 +580,36 @@ function copySourceIntoTextArea() {
 			li = li.nextSibling
 		}
 		textarea.value = strings.join('\n')
+		// Strip CSS-checking wrapper if it leaked into the source display.
+		// The server wraps CSS input in an HTML document with <style>; the
+		// source display normally strips this, but guard against stacking.
+		var cssProlog = "<!DOCTYPE html><html lang=\'\'><title>s</title><style>"
+		var cssEpilog = "</style>"
+		var v = textarea.value
+		if (v.indexOf(cssProlog) === 0
+				&& v.lastIndexOf(cssEpilog) === v.length - cssEpilog.length) {
+			v = v.substring(cssProlog.length, v.length - cssEpilog.length)
+			if (v.charAt(0) === '\n') v = v.substring(1)
+			if (v.charAt(v.length - 1) === '\n') v = v.substring(0, v.length - 1)
+			textarea.value = v
+		}
+	}
+}
+
+function hideSourceIfNotRequested() {
+	var showSourceCheckbox = document.getElementById("showsource")
+	if (showSourceCheckbox && !showSourceCheckbox.checked) {
+		var sourceHeading = document.getElementById('source')
+		if (sourceHeading) {
+			var sourceList = sourceHeading.nextSibling
+			while (sourceList && sourceList.nodeType != 1) {
+				sourceList = sourceList.nextSibling
+			}
+			if (sourceList && sourceList.className == 'source') {
+				sourceHeading.style.display = 'none'
+				sourceList.style.display = 'none'
+			}
+		}
 	}
 }
 
@@ -627,6 +677,11 @@ if (document.getElementById) {
 	boot()
 }
 
+/**
+ * Categorizes a validation message as 'css', 'i18n', or 'html'.
+ * NOTE: This function is duplicated in site/message-category.js for unit testing.
+ * If you modify this function, update message-category.js to match.
+ */
 function getMessageCategory(messageText) {
 	// CSS validation errors (always prefixed with "CSS:")
 	if (/^CSS:/.test(messageText)) {
@@ -637,6 +692,8 @@ function getMessageCategory(messageText) {
 	if (/\b(encoding|charset|UTF-8|windows-\d+|iso-\d+|Content-Language)\b/i.test(messageText) ||
 		/appears to be written in/i.test(messageText) ||
 		/\b(lang|dir)=/i.test(messageText) ||
+		/"lang"/.test(messageText) ||
+		/"dir"/.test(messageText) ||
 		/Unicode Normalization/.test(messageText)) {
 		return 'i18n'
 	}
